// -*- mode: js; js-indent-level: 4; indent-tabs-mode: nil -*-
/*
  Copyright (C) 2012-2019  Dieter Schärf <ds@dschaerf.de>

  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

const Clutter = imports.gi.Clutter;
const GMenu = imports.gi.GMenu;
const Lang = imports.lang;
const Shell = imports.gi.Shell;
const St = imports.gi.St;

const ExtensionUtils = imports.misc.extensionUtils;
const Main = imports.ui.main;
const PanelMenu = imports.ui.panelMenu;
const PopupMenu = imports.ui.popupMenu;

const Me = ExtensionUtils.getCurrentExtension();

const Convenience = Me.imports.convenience;

const _ = Convenience.getTranslations();

const KEY_SHOW_APPLICATIONS = 'show-applications';

const PANELMENU_BUTTON_ICON = 1;
const PANELMENU_BUTTON_TEXT = 2;

const ApplicationMenuItem = new Lang.Class({
    Name: 'PanelManager.ApplicationMenuItem',
    Extends: PopupMenu.PopupBaseMenuItem,

    _init: function(app) {
        this.parent();
        this._icon = new St.Icon({ style_class: 'popup-menu-icon' });
        this._icon.gicon = app.get_app_info().get_icon();
        this.actor.add_child(this._icon);
        this._label = new St.Label({ text: app.get_name(), y_align: Clutter.ActorAlign.CENTER });
        this.actor.add_child(this._label);
        this.actor.label_actor = this._label;
        this.connect('activate', Lang.bind(this, function() {
            app.activate();
            Main.overview.hide();
        }));
    }
});

const ApplicationsMenuButton = new Lang.Class({
    Name: 'PanelManager.ApplicationsMenuButton',
    Extends: PanelMenu.Button,

    _init: function() {
        this.parent(1.0, _("Applications"), false);
        this.menu.actor.add_style_class_name('dschaerf-menu');
        this._hbox = new St.BoxLayout({ style_class: 'panel-status-menu-box' });
        this._icon = new St.Icon({ icon_name: 'view-app-grid-symbolic', style_class: 'system-status-icon' });
        this._hbox.add_child(this._icon);
        this._label = new St.Label({ text: _("Applications"), y_expand: true, y_align: Clutter.ActorAlign.CENTER });
        this._hbox.add_child(this._label);
        this._hbox.add_child(PopupMenu.arrowIcon(St.Side.BOTTOM));
        this.actor.add_child(this._hbox);
        this._settingsChangedId = Convenience.Settings.connect('changed::'+KEY_SHOW_APPLICATIONS, Lang.bind(this, this._onShowApplications));
        this._appSystem = Shell.AppSystem.get_default();
        this._appSystemChangedId = [
            this._appSystem.connect('app-state-changed', Lang.bind(this, this._updateApplications)),
            this._appSystem.connect('installed-changed', Lang.bind(this, this._updateApplications))
        ];
        this._onShowApplications();
        this._updateApplications();
        this.actor.connect('destroy', Lang.bind(this, function() {
            this._appSystemChangedId.forEach(function(id) {
                this._appSystem.disconnect(id);
            }, this);
            Convenience.Settings.disconnect(this._settingsChangedId);
        }));
    },

    _getCategory: function(appList, appDir) {
        let iter = appDir.iter();
        let nextType;
        while ((nextType = iter.next()) !== GMenu.TreeItemType.INVALID) {
            if (nextType === GMenu.TreeItemType.ENTRY) {
                let entry = iter.get_entry();
                let app = this._appSystem.lookup_app(entry.get_desktop_file_id());
                if (entry.get_app_info().should_show()) appList.push(app);
            } else if (nextType === GMenu.TreeItemType.DIRECTORY) {
                if (!appDir.get_is_nodisplay()) this._getCategory(appList, iter.get_directory());
            }
        }
    },

    _onShowApplications: function() {
        let showApplications = Convenience.Settings.get_enum(KEY_SHOW_APPLICATIONS);
        this._icon.visible = showApplications !== PANELMENU_BUTTON_TEXT;
        this._label.visible = showApplications !== PANELMENU_BUTTON_ICON;
    },

    _updateApplications: function() {
        this.menu.removeAll();
        let tree = new GMenu.Tree({ menu_basename: 'applications.menu' });
        tree.load_sync();
        let root = tree.get_root_directory();
        let iter = root.iter();
        let nextType;
        while ((nextType = iter.next()) !== GMenu.TreeItemType.INVALID) {
            switch (nextType) {
                case GMenu.TreeItemType.SEPARATOR:
                    this.menu.addMenuItem(new PopupMenu.PopupSeparatorMenuItem());
                    break;
                case GMenu.TreeItemType.ENTRY:
                    let entry = iter.get_entry();
                    let app = this._appSystem.lookup_app(entry.get_desktop_file_id());
                    if (entry.get_app_info().should_show()) this.menu.addMenuItem(new ApplicationMenuItem(app));
                    break;
                case GMenu.TreeItemType.DIRECTORY:
                    let appDir = iter.get_directory();
                    if (appDir.get_is_nodisplay()) continue;
                    let item = new PopupMenu.PopupSubMenuMenuItem(appDir.get_name(), true);
                    item.icon.gicon = appDir.get_icon();
                    let appList = [];
                    this._getCategory(appList, appDir);
                    appList.sort(function(a, b) {
                        return a.compare_by_name(b);
                    });
                    appList.forEach(function(app) {
                        item.menu.addMenuItem(new ApplicationMenuItem(app));
                    }, this);
                    if (item.menu.length > 0) this.menu.addMenuItem(item);
                    break;
                default: // Do nothing
            }
        }
    }
});

function getExtension() {
    return Main.panel.addToStatusArea('dschaerfApplications', new ApplicationsMenuButton(), 1, 'left');
}
