// -*- mode: js; js-indent-level: 4; indent-tabs-mode: nil -*-
/*
  Copyright (C) 2012-2019  Dieter Schärf <ds@dschaerf.de>

  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

const Clutter = imports.gi.Clutter;
const Lang = imports.lang;
const Shell = imports.gi.Shell;
const St = imports.gi.St;

const ExtensionUtils = imports.misc.extensionUtils;
const Main = imports.ui.main;
const PanelMenu = imports.ui.panelMenu;
const PopupMenu = imports.ui.popupMenu;

const Me = ExtensionUtils.getCurrentExtension();

const Convenience = Me.imports.convenience;

const _ = Convenience.getTranslations();

const FAVORITES_ALL = 1;

const KEY_FAVORITE_APPS = 'favorite-apps';
const KEY_FAVORITES_SCOPE = 'favorites-scope';
const KEY_SHOW_FAVORITES = 'show-favorites';

const PANELMENU_BUTTON_ICON = 1;
const PANELMENU_BUTTON_TEXT = 2;

const FavoriteMenuItem = new Lang.Class({
    Name: 'PanelManager.FavoriteMenuItem',
    Extends: PopupMenu.PopupBaseMenuItem,

    _init: function(app) {
        this.parent();
        this._icon = new St.Icon({ style_class: 'popup-menu-icon' });
        this._icon.gicon = app.get_app_info().get_icon();
        this.actor.add_child(this._icon);
        this._label = new St.Label({ text: app.get_name(), y_align: Clutter.ActorAlign.CENTER });
        this.actor.add_child(this._label);
        this.actor.label_actor = this._label;
        this.connect('activate', Lang.bind(this, function(event) {
            app.activate();
            Main.overview.hide();
        }));
    }
});

const FavoritesMenuButton = new Lang.Class({
    Name: 'PanelManager.FavoritesMenuButton',
    Extends: PanelMenu.Button,

    _init: function() {
        this.parent(1.0, _("Favorites"), false);
        this.menu.actor.add_style_class_name('dschaerf-menu');
        this._hbox = new St.BoxLayout({ style_class: 'panel-status-menu-box' });
        this._icon = new St.Icon({ icon_name: 'emblem-favorite-symbolic', style_class: 'system-status-icon' });
        this._hbox.add_child(this._icon);
        this._label = new St.Label({ text: _("Favorites"), y_expand: true, y_align: Clutter.ActorAlign.CENTER });
        this._hbox.add_child(this._label);
        this._hbox.add_child(PopupMenu.arrowIcon(St.Side.BOTTOM));
        this.actor.add_child(this._hbox);
        this._settingsChangedId = [
            Convenience.Settings.connect('changed::'+KEY_FAVORITE_APPS, Lang.bind(this, this._updateFavorites)),
            Convenience.Settings.connect('changed::'+KEY_FAVORITES_SCOPE, Lang.bind(this, this._updateFavorites)),
            Convenience.Settings.connect('changed::'+KEY_SHOW_FAVORITES, Lang.bind(this, this._onShowFavorites))
        ];
        this._appSystem = Shell.AppSystem.get_default();
        this._appSystemChangedId = this._appSystem.connect('installed-changed', Lang.bind(this, this._updateFavorites));
        this._globalSettingsChangedId = global.settings.connect('changed::'+KEY_FAVORITE_APPS, Lang.bind(this, this._updateFavorites));
        this._onShowFavorites();
        this._updateFavorites();
        this.actor.connect('destroy', Lang.bind(this, function() {
            global.settings.disconnect(this._globalSettingsChangedId);
            this._appSystem.disconnect(this._appSystemChangedId);
            this._settingsChangedId.forEach(function(id) {
                Convenience.Settings.disconnect(id);
            }, this);
        }));
    },

    _onShowFavorites: function() {
        let showFavorites = Convenience.Settings.get_enum(KEY_SHOW_FAVORITES);
        this._icon.visible = showFavorites !== PANELMENU_BUTTON_TEXT;
        this._label.visible = showFavorites !== PANELMENU_BUTTON_ICON;
    },

    _updateFavorites: function() {
        this.menu.removeAll();
        let appList = [];
        let favorites = Convenience.Settings.get_strv(KEY_FAVORITE_APPS)
        favorites.forEach(function(desktopFile) {
            let app = this._appSystem.lookup_app(desktopFile);
            if (app) appList.push(app);
        }, this);
        if (Convenience.Settings.get_enum(KEY_FAVORITES_SCOPE) === FAVORITES_ALL) {
            let favoriteApps = global.settings.get_strv(KEY_FAVORITE_APPS);
            favoriteApps.forEach(function(desktopFile) {
                let app = this._appSystem.lookup_app(desktopFile);
                if (app && appList.indexOf(app) < 0) appList.push(app);
            }, this);
        }
        appList.sort(function(a, b) {
            return a.compare_by_name(b);
        });
        appList.forEach(function(app) {
            this.menu.addMenuItem(new FavoriteMenuItem(app));
        }, this);
        this.actor.visible = this.menu.numMenuItems > 0;
    }
});

function getExtension() {
    return Main.panel.addToStatusArea('dschaerfFavorites', new FavoritesMenuButton(), 1, 'left');
}
