// -*- mode: js; js-indent-level: 4; indent-tabs-mode: nil -*-
/*
  Copyright (C) 2012-2019  Dieter Schärf <ds@dschaerf.de>

  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

const Clutter = imports.gi.Clutter;
const Gio = imports.gi.Gio;
const GLib = imports.gi.GLib;
const Gtk = imports.gi.Gtk;
const Lang = imports.lang;
const Mainloop = imports.mainloop;
const Shell = imports.gi.Shell;
const Signals = imports.signals;
const St = imports.gi.St;

const ExtensionUtils = imports.misc.extensionUtils;
const Main = imports.ui.main;
const PanelMenu = imports.ui.panelMenu;
const PopupMenu = imports.ui.popupMenu;
const ShellMountOperation = imports.ui.shellMountOperation;

const Me = ExtensionUtils.getCurrentExtension();

const Convenience = Me.imports.convenience;
const UserDialog = Me.imports.userDialog;

const _ = Convenience.getTranslations();

const BOOKMARKS_MAINMENU = 0;

const KEY_BOOKMARKS_POSITION = 'bookmarks-position';
const KEY_SHOW_PLACES = 'show-places';

const PANELMENU_BUTTON_ICON = 1;
const PANELMENU_BUTTON_TEXT = 2;

const PlaceInfo = new Lang.Class({
    Name: 'PlaceInfo',

    _init: function(category, file, label) {
        this._category = category;
        this.file = file;
        this.icon = this._getIcon();
        this.label = label || this._getLabel();
        if (this.isTrash()) {
            this._trashMonitor = this.file.monitor_directory(0, null);
            this._trashMonitorChangedId = this._trashMonitor.connect('changed', Lang.bind(this, this._onTrashChanged));
            this._onTrashChanged();
        }
    },

    _getIcon: function() {
        try {
            let info = this.file.query_info('standard::icon', 0, null);
            return info.get_icon();
        } catch(e) {
            if (e instanceof Gio.IOErrorEnum) {
                switch (this._category) {
                    case 'computer':
                        return new Gio.ThemedIcon({ name: 'computer' });
                    case 'file-system':
                        return new Gio.ThemedIcon({ name: 'folder' });
                    case 'network-workgroup':
                        return new Gio.ThemedIcon({ name: 'network-workgroup' });
                    case 'volumes-device':
                        return new Gio.ThemedIcon({ name: 'drive-harddisk' });
                    case 'volumes-network':
                        return new Gio.ThemedIcon({ name: 'folder-remote' });
                    default:
                        if (!this.file.is_native()) return new Gio.ThemedIcon({ name: 'folder-remote' });
                        return new Gio.ThemedIcon({ name: 'folder' });
                }
            } else {
                throw e;
            }
        }
    },

    _getLabel: function() {
        try {
            let info = this.file.query_info('standard::display-name', 0, null)
            return info.get_display_name();
        } catch(e) {
            if (e instanceof Gio.IOErrorEnum) return this.file.get_basename();
            throw e;
        }
    },

    _onTrashChanged: function() {
        this.icon = this._getIcon();
        this.emit('trash-changed');
    },

    destroy: function() {
        if (this._trashMonitorChangedId) this._trashMonitor.disconnect(this._trashMonitorChangedId);
    },

    isRemovable: function() {
        return false;
    },

    isTrash: function() {
        return this._category === 'trash';
    },

    launch: function() {
        let launchContext = global.create_app_launch_context(0, -1);
        try {
            Gio.AppInfo.launch_default_for_uri(this.file.get_uri(), launchContext);
        } catch(e) {
            if (e.matches(Gio.IOErrorEnum, Gio.IOErrorEnum.NOT_MOUNTED)) {
                try {
                    this.file.mount_enclosing_volume(0, null, null, function(file, result) {
                        file.mount_enclosing_volume_finish(result);
                        Gio.AppInfo.launch_default_for_uri(file.get_uri(), launchContext);
                    });
                } catch(e) {
                    Main.notifyError(_("Failed to launch '%s'").format(this.label), e.message);
                }
            } else {
                Main.notifyError(_("Failed to launch '%s'").format(this.label), e.message);
            }
        }
    }
});
Signals.addSignalMethods(PlaceInfo.prototype);

const PlaceMountInfo = new Lang.Class({
    Name: 'PlaceMountInfo',
    Extends: PlaceInfo,

    _init: function(category, mount) {
        this._mount = mount;
        this.parent(category, mount.get_root(), mount.get_name());
    },

    _getIcon: function() {
        return this._mount.get_icon();
    },

    _removeFinish: function(mount, result) {
        try {
            if (mount.can_eject()) {
                mount.eject_with_operation_finish(result);
            } else {
                mount.unmount_with_operation_finish(result);
            }
        } catch (e) {
            Main.notifyError(_("Failed to unmount '%s'").format(this.label), e.message);
        }
    },

    isRemovable: function() {
        if (!this._mount.can_eject() && !this._mount.can_unmount()) return false;
        if (this._mount.is_shadowed()) return false;
        let volume = this._mount.get_volume();
        if (volume === null) return true;
        return volume.get_identifier('class') !== 'network';
    },

    remove: function() {
        if (!this.isRemovable()) return;
        let mountOp = new ShellMountOperation.ShellMountOperation(this._mount);
        if (this._mount.can_eject()) {
            this._mount.eject_with_operation(Gio.MountUnmountFlags.NONE, mountOp.mountOp, null, Lang.bind(this, this._removeFinish));
        } else {
            this._mount.unmount_with_operation(Gio.MountUnmountFlags.NONE, mountOp.mountOp, null, Lang.bind(this, this._removeFinish));
        }
    }
});

const PlacesManager = new Lang.Class({
    Name: 'PlacesManager',

    _init: function() {
        this._places = {};
        this._bookmarksFile = this._getBookmarksFile();
        this._bookmarksMonitor = null;
        this._bookmarksTimeoutId = 0;
        if (this._bookmarksFile) {
            this._bookmarksMonitor = this._bookmarksFile.monitor_file(Gio.FileMonitorFlags.NONE, null);
            this._bookmarksMonitorChangedId = this._bookmarksMonitor.connect('changed', Lang.bind(this, function () {
                if (this._bookmarksTimeoutId > 0) return;
                this._bookmarksTimeoutId = Mainloop.timeout_add(100, Lang.bind(this, function () {
                    this._bookmarksTimeoutId = 0;
                    this._updateBookmarks();
                    return false;
                }));
            }));
        }
        this._gioSettings = new Gio.Settings({ schema_id: 'org.gnome.desktop.background' });
        this._gioSettingsChangedId = this._gioSettings.connect('changed::show-desktop-icons', Lang.bind(this, this._updateBookmarks));
        this._volumeMonitor = Gio.VolumeMonitor.get();
        this._volumeMonitorChangedId = [
            this._volumeMonitor.connect('drive-changed', Lang.bind(this, this._updateVolumes)),
            this._volumeMonitor.connect('drive-connected', Lang.bind(this, this._updateVolumes)),
            this._volumeMonitor.connect('drive-disconnected', Lang.bind(this, this._updateVolumes)),
            this._volumeMonitor.connect('mount-added', Lang.bind(this, this._updateVolumes)),
            this._volumeMonitor.connect('mount-changed', Lang.bind(this, this._updateVolumes)),
            this._volumeMonitor.connect('mount-removed', Lang.bind(this, this._updateVolumes)),
            this._volumeMonitor.connect('volume-added', Lang.bind(this, this._updateVolumes)),
            this._volumeMonitor.connect('volume-changed', Lang.bind(this, this._updateVolumes)),
            this._volumeMonitor.connect('volume-removed',Lang.bind(this, this._updateVolumes))
        ];
        this._updateBookmarks();
        this._updateVolumes();
    },

    _addMount: function(category, mount) {
        this._places[category].push(new PlaceMountInfo(category, mount));
    },

    _getBookmarksFile: function() {
        let files = [
            GLib.build_filenamev([ GLib.get_user_config_dir(), 'gtk-3.0', 'bookmarks' ]),
            GLib.build_filenamev([ GLib.get_home_dir(), '.gtk-bookmarks' ]),
        ];
        for (let i = 0; i < files.length; i++) {
            let file = files[i];
            if (GLib.file_test(file, GLib.FileTest.EXISTS)) return Gio.File.new_for_path(file);
        }
        return null;
    },

    _sortCategory: function(category) {
        this._places[category].sort(function(a, b) {
            return GLib.utf8_collate(a.label, b.label);
        });
    },

    _updateBookmarks: function() {
        this._places['bookmarks-custom'] = [];
        this._places['bookmarks-default'] = [];
        this._places['bookmarks-special'] = [];
        this._places['bookmarks-special'].push(new PlaceInfo('default', Gio.File.new_for_path(GLib.get_home_dir()), _("Home Folder")));
        if (this._gioSettings.get_boolean('show-desktop-icons')) {
            let bookmark = GLib.get_user_special_dir(GLib.UserDirectory.DIRECTORY_DESKTOP);
            if (GLib.file_test(bookmark, GLib.FileTest.EXISTS)) this._places['bookmarks-special'].push(new PlaceInfo('default', Gio.File.new_for_path(bookmark), null));
        }
        for (let i = 0; i < GLib.UserDirectory.N_DIRECTORIES; i++) {
            if (i === GLib.UserDirectory.DIRECTORY_DESKTOP) continue;
            let bookmark = GLib.get_user_special_dir(i);
            if (GLib.file_test(bookmark, GLib.FileTest.EXISTS)) this._places['bookmarks-default'].push(new PlaceInfo('default', Gio.File.new_for_path(bookmark), null));
        }
        if (this._bookmarksFile) {
            let content = Shell.get_file_contents_utf8_sync(this._bookmarksFile.get_path());
            let lines = content.split('\n');
            lines.forEach(function(line) {
                let components = line.split(' ');
                let component = components[0];
                if (component) {
                    let bookmark = Gio.File.new_for_uri(component);
                    let duplicate = false;
                    this._places['bookmarks-special'].forEach(function(item) {
                        if (bookmark.equal(item.file)) duplicate = true;
                    });
                    if (!duplicate) this._places['bookmarks-default'].forEach(function(item) {
                        if (bookmark.equal(item.file)) duplicate = true;
                    });
                    if (!duplicate) this._places['bookmarks-custom'].push(new PlaceInfo('default', bookmark, components[1] || ''));
                }
            }, this);
        }
        this._sortCategory('bookmarks-custom');
        this._sortCategory('bookmarks-default');
        this.emit('bookmarks-updated');
    },

    _updateVolumes: function() {
        this._places['volumes-device'] = [];
        this._places['volumes-network'] = [];
        this._places['volumes-special'] = [];
        this._places['volumes-special'].push(new PlaceInfo('file-system', Gio.File.new_for_uri('file:///'), _("File System")));
        this._places['volumes-special'].push(new PlaceInfo('trash', Gio.File.new_for_uri('trash:///'), _("Trash")));
        this._places['volumes-special'].push(new PlaceInfo('computer', Gio.File.new_for_uri('computer:///'), _("Computer")));
        this._places['volumes-special'].push(new PlaceInfo('network-workgroup', Gio.File.new_for_uri('network:///'), _("Network")));
        let drives = this._volumeMonitor.get_connected_drives();
        drives.forEach(function(drive) {
            let volumes = drive.get_volumes();
            volumes.forEach(function(volume) {
                let identifier = volume.get_identifier('class');
                let category = identifier && identifier.indexOf('network') >= 0 ? 'volumes-network' : 'volumes-device';
                let mount = volume.get_mount();
                if (mount !== null) this._addMount(category, mount);
            }, this);
        }, this);
        let volumes = this._volumeMonitor.get_volumes();
        volumes.forEach(function(volume) {
            if (volume.get_drive() === null) {
                let identifier = volume.get_identifier('class');
                let category = identifier && identifier.indexOf('network') >= 0 ? 'volumes-network' : 'volumes-device';
                let mount = volume.get_mount();
                if (mount !== null) this._addMount(category, mount);
            }
        }, this);
        let mounts = this._volumeMonitor.get_mounts();
        mounts.forEach(function(mount) {
            if (!mount.is_shadowed() && !mount.get_volume()) {
                let location = mount.get_default_location();
                let category = location.is_native() ? 'volumes-device' : 'volumes-network';
                this._addMount(category, mount);
            }
        }, this);
        this._sortCategory('volumes-device');
        this._sortCategory('volumes-network');
        this.emit('volumes-updated');
    },

    destroy: function() {
        this._volumeMonitorChangedId.forEach(function(connection) {
            this._volumeMonitor.disconnect(connection);
        }, this);
        this._gioSettings.disconnect(this._gioSettingsChangedId);
        if (this._bookmarksTimeoutId) Mainloop.source_remove(this._bookmarksTimeoutId);
        if (this._bookmarksMonitor) this._bookmarksMonitor.disconnect(this._bookmarksMonitorChangedId);
        for (let category in this._places) {
            this._places[category].forEach(function(info) {
                info.destroy();
            });
        }
    },

    get: function(category) {
        return this._places[category];
    }
});
Signals.addSignalMethods(PlacesManager.prototype);

const PlaceMenuItem = new Lang.Class({
    Name: 'PanelManager.PlaceMenuItem',
    Extends: PopupMenu.PopupBaseMenuItem,

    _init: function(text, iconName) {
        this.parent();
        this._icon = new St.Icon({ style_class: 'popup-menu-icon' });
        if (iconName) this._icon.icon_name = iconName;
        this.actor.add_child(this._icon);
        this._label = new St.Label({ text: text, y_align: Clutter.ActorAlign.CENTER });
        this.actor.add_child(this._label);
        this.actor.label_actor = this._label;
    },

    setIcon: function(icon) {
        this._icon.gicon = icon;
    }
});

const PlaceLaunchMenuItem = new Lang.Class({
    Name: 'PanelManager.PlaceLaunchMenuItem',
    Extends: PlaceMenuItem,

    _init: function(place) {
        this.parent(place.label, null);
        this.setIcon(place.icon);
        this._place = place;
        if (place.isRemovable()) this._addButton(place);
        if (place.isTrash()) place.connect('trash-changed', Lang.bind(this, function() {
            this.setIcon(place.icon);
        }));
        this.connect('activate', Lang.bind(this, function() {
            place.launch();
            Main.overview.hide();
        }));
    },

    _addButton: function(place) {
        this.actor.add(new St.Bin(), { expand: true });
        this._button = new St.Button({ style_class: 'dschaerf-button system-menu-action' });
        this._button.set_child(new St.Icon({ icon_name: 'media-eject-symbolic' }));
        this._button.connect('clicked', Lang.bind(this, function() {
            let topMenu = this._getTopMenu();
            if (topMenu.isOpen) topMenu.close();
            let dialog = new UserDialog.ModalUserDialog('EjectVolume', place.label.toUpperCase());
            dialog.connect('eject-volume', Lang.bind(this, function() {
                place.remove();
            }));
            dialog.open();
            Main.overview.hide();
        }));
        this.actor.add_child(this._button, { align: St.Align.END });
    }
});

const PlaceRecentDocumentMenuItem = new Lang.Class({
    Name: 'PanelManager.PlaceRecentDocumentMenuItem',
    Extends: PlaceMenuItem,

    _init: function(file) {
        this.parent(file.get_display_name(), null);
        this.setIcon(Gio.content_type_get_icon(file.get_mime_type()));
        this._file = file;
        this.connect('activate', Lang.bind(this, function() {
            Gio.AppInfo.launch_default_for_uri(file.get_uri(), global.create_app_launch_context(0, -1));
            Main.overview.hide();
        }));
    }
});

const PlacesMenuButton = new Lang.Class({
    Name: 'PanelManager.PlacesMenuButton',
    Extends: PanelMenu.Button,

    _init: function() {
        this.parent(1.0, _("Places"), false);
        this.menu.actor.add_style_class_name('dschaerf-menu');
        this._hbox = new St.BoxLayout({ style_class: 'panel-status-menu-box' });
        this._icon = new St.Icon({ icon_name: 'folder-symbolic', style_class: 'system-status-icon' });
        this._hbox.add_child(this._icon);
        this._label = new St.Label({ text: _("Places"), y_expand: true, y_align: Clutter.ActorAlign.CENTER });
        this._hbox.add_child(this._label);
        this._hbox.add_child(PopupMenu.arrowIcon(St.Side.BOTTOM));
        this.actor.add_child(this._hbox);
        this._itemBookmarks = new PopupMenu.PopupMenuSection();
        this.menu.addMenuItem(this._itemBookmarks);
        this.menu.addMenuItem(new PopupMenu.PopupSeparatorMenuItem());
        this._itemVolumes = new PopupMenu.PopupMenuSection();
        this.menu.addMenuItem(this._itemVolumes);
        this._separatorRecentDocuments = new PopupMenu.PopupSeparatorMenuItem();
        this.menu.addMenuItem(this._separatorRecentDocuments);
        this._itemRecentDocuments = new PopupMenu.PopupMenuSection();
        this._itemRecentDocumentsClear = new PlaceMenuItem(_("Clear the Recent Documents list…"), 'edit-clear-all');
        this._itemRecentDocumentsClear.connect('activate', Lang.bind(this, this._onClearRecentDocuments));
        this._itemRecentDocuments.addMenuItem(this._itemRecentDocumentsClear);
        this._itemRecentDocumentsContent = new PopupMenu.PopupSubMenuMenuItem(_("Recent Documents"), true);
        this._itemRecentDocumentsContent.icon.icon_name = 'document-open-recent';
        this._itemRecentDocuments.addMenuItem(this._itemRecentDocumentsContent);
        this.menu.addMenuItem(this._itemRecentDocuments);
        this._settingsChangedId = [
            Convenience.Settings.connect('changed::'+KEY_BOOKMARKS_POSITION, Lang.bind(this, this._updateBookmarks)),
            Convenience.Settings.connect('changed::'+KEY_SHOW_PLACES, Lang.bind(this, this._onShowPlaces))
        ];
        this._placesManager = new PlacesManager();
        this._placesManagerChangedId = [
            this._placesManager.connect('bookmarks-updated', Lang.bind(this, this._updateBookmarks)),
            this._placesManager.connect('volumes-updated', Lang.bind(this, this._updateVolumes))
        ];
        this._recentManager = new Gtk.RecentManager();
        this.menu.connect('open-state-changed', Lang.bind(this, function(menu, isOpen) {
            if (isOpen) this._updateRecentDokuments();
        }));
        this._onShowPlaces();
        this._updateBookmarks();
        this._updateVolumes();
        this.actor.connect('destroy', Lang.bind(this, function() {
            this._placesManagerChangedId.forEach(function(id) {
                this._placesManager.disconnect(id);
            }, this);
            this._settingsChangedId.forEach(function(id) {
                Convenience.Settings.disconnect(id);
            });
            this._placesManager.destroy();
        }));
    },

    _onClearRecentDocuments: function() {
        let dialog = new UserDialog.ModalUserDialog('ClearRecentDocuments');
        dialog.connect('clear-recent-documents', Lang.bind(this, function() {
            this._recentManager.purge_items();
            this._separatorRecentDocuments.actor.hide();
            this._itemRecentDocuments.actor.hide();
        }));
        dialog.open();
        Main.overview.hide();
    },

    _onShowPlaces: function() {
        let showPlaces = Convenience.Settings.get_enum(KEY_SHOW_PLACES);
        this._icon.visible = showPlaces !== PANELMENU_BUTTON_TEXT;
        this._label.visible = showPlaces !== PANELMENU_BUTTON_ICON;
    },

    _updateBookmarks: function() {
        this._itemBookmarks.removeAll();
        let bookmarksSpecial = this._placesManager.get('bookmarks-special');
        bookmarksSpecial.forEach(function(bookmark) {
            this._itemBookmarks.addMenuItem(new PlaceLaunchMenuItem(bookmark));
        }, this);
        let bookmarksDefault = this._placesManager.get('bookmarks-default');
        if (bookmarksDefault.length > 0) {
            if (Convenience.Settings.get_enum(KEY_BOOKMARKS_POSITION) === BOOKMARKS_MAINMENU) {
                bookmarksDefault.forEach(function(bookmark) {
                    this._itemBookmarks.addMenuItem(new PlaceLaunchMenuItem(bookmark));
                }, this);
            } else {
                let item = new PopupMenu.PopupSubMenuMenuItem(_("Bookmarks"), true);
                item.icon.icon_name = 'user-bookmarks';
                bookmarksDefault.forEach(function(bookmark) {
                    item.menu.addMenuItem(new PlaceLaunchMenuItem(bookmark));
                });
                this._itemBookmarks.addMenuItem(item);
            }
        }
        let bookmarksCustom = this._placesManager.get('bookmarks-custom');
        if (bookmarksCustom.length > 0) {
            let item = new PopupMenu.PopupSubMenuMenuItem(_("Bookmarks (user defined)"), true);
            item.icon.icon_name = 'user-bookmarks';
            bookmarksCustom.forEach(function(bookmark) {
                item.menu.addMenuItem(new PlaceLaunchMenuItem(bookmark));
            });
            this._itemBookmarks.addMenuItem(item);
        }
    },

    _updateRecentDokuments: function() {
        this._itemRecentDocumentsContent.menu.removeAll();
        if (this._recentManager.size > 0) {
            let recentItems = this._recentManager.get_items();
            recentItems.sort(function(a, b) {
                return GLib.utf8_collate(a.get_display_name(), b.get_display_name());
            });
            recentItems.forEach(function(item) {
                if (GLib.file_test(item.get_uri_display(), GLib.FileTest.IS_REGULAR)) this._itemRecentDocumentsContent.menu.addMenuItem(new PlaceRecentDocumentMenuItem(item));
            }, this);
        }
        let visible = this._itemRecentDocumentsContent.menu.numMenuItems > 0;
        this._separatorRecentDocuments.actor.visible = visible;
        this._itemRecentDocuments.actor.visible = visible;
    },

    _updateVolumes: function() {
        this._itemVolumes.removeAll();
        let volumesSpecial = this._placesManager.get('volumes-special');
        volumesSpecial.forEach(function(volume) {
            this._itemVolumes.addMenuItem(new PlaceLaunchMenuItem(volume));
        }, this);
        let volumesDevice = this._placesManager.get('volumes-device');
        if (volumesDevice.length > 0) {
            this._itemRemovableDevices = new PopupMenu.PopupSubMenuMenuItem(_("Removable Devices"), true);
            this._itemRemovableDevices.icon.icon_name = 'media-removable';
            volumesDevice.forEach(function(volume) {
                this._itemRemovableDevices.menu.addMenuItem(new PlaceLaunchMenuItem(volume));
            }, this);
            this._itemVolumes.addMenuItem(this._itemRemovableDevices);
        }
        let volumesNetwork = this._placesManager.get('volumes-network');
        if (volumesNetwork.length > 0) {
            this._itemVolumesNetwork = new PopupMenu.PopupSubMenuMenuItem(_("Network Places"), true);
            this._itemVolumesNetwork.icon.icon_name = 'network';
            volumesNetwork.forEach(function(volume) {
                this._itemVolumesNetwork.menu.addMenuItem(new PlaceLaunchMenuItem(volume));
            }, this);
            this._itemVolumes.addMenuItem(this._itemVolumesNetwork);
        }
    }
});

function getExtension() {
    return Main.panel.addToStatusArea('dschaerfPlaces', new PlacesMenuButton(), 1, 'left');
}
