// -*- mode: js; js-indent-level: 4; indent-tabs-mode: nil -*-
/*
  Copyright (C) 2012-2019  Dieter Schärf <ds@dschaerf.de>

  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

const Clutter = imports.gi.Clutter;
const Gio = imports.gi.Gio;
const GLib = imports.gi.GLib;
const Lang = imports.lang;
const St = imports.gi.St;

const ExtensionUtils = imports.misc.extensionUtils;
const Main = imports.ui.main;
const PanelMenu = imports.ui.panelMenu;
const PopupMenu = imports.ui.popupMenu;

const Me = ExtensionUtils.getCurrentExtension();

const Convenience = Me.imports.convenience;
const UserDialog = Me.imports.userDialog;

const _ = Convenience.getTranslations();

const TRASH_CONTENT_MAX = 20;
const TRASH_PATH = 'trash:///';
const TRASH_URI = Gio.File.new_for_uri(TRASH_PATH);

const TrashMenuItem = new Lang.Class({
    Name: 'PanelTools.TrashMenuItem',
    Extends: PopupMenu.PopupBaseMenuItem,

    _init: function(text, iconName) {
        this.parent();
        this._icon = new St.Icon({ style_class: 'popup-menu-icon' });
        this.actor.add_child(this._icon);
        this._label = new St.Label({ text: text, y_align: Clutter.ActorAlign.CENTER });
        this.actor.add_child(this._label);
        this.actor.label_actor = this._label;
        if (iconName) this._icon.icon_name = iconName;
    }
});

const TrashEntryMenuItem = new Lang.Class({
    Name: 'PanelTools.TrashEntryMenuItem',
    Extends: TrashMenuItem,

    _init: function(trashEntry) {
        this.parent(trashEntry.get_name(), null);
        this._icon.gicon = trashEntry.get_icon();
        this.actor.add(new St.Bin(), { expand: true });
        this._button = new St.Button({ style_class: 'dschaerf-button system-menu-action' });
        this._button.set_child(new St.Icon({ icon_name: 'edit-delete-symbolic' }));
        this._button.connect('clicked', Lang.bind(this, function(event) {
            if (this._getTopMenu().isOpen) this._getTopMenu().close();
            let dialog = new UserDialog.ModalUserDialog('TrashEntryDelete', trashEntry.get_display_name());
            dialog.connect('trash-entry-delete', Lang.bind(this, function() {
                TRASH_URI.get_child(trashEntry.get_name()).delete(null);
            }));
            dialog.open();
            Main.overview.hide();
        }));
        this.actor.add_child(this._button);
        this.connect('activate', Lang.bind(this, function() {
            Gio.AppInfo.launch_default_for_uri(TRASH_URI.get_child(trashEntry.get_name()).get_uri(), global.create_app_launch_context(0, -1));
            Main.overview.hide();
        }));
    }
});

const TrashSubMenuMenuItem = new Lang.Class({
    Name: 'PanelManager.TrashSubMenuMenuItem',
    Extends: PopupMenu.PopupSubMenuMenuItem,

    _init: function(text, iconName) {
        this.parent(text, true);
        this.icon.icon_name = iconName;
    },

    setStatus: function(text) {
        if (!this._statusLabel) {
            this._statusLabel = new St.Label({ y_expand: true, y_align: Clutter.ActorAlign.CENTER });
            let children = this.actor.get_n_children();
            this.actor.insert_child_at_index(this._statusLabel, children - 1);
        }
        this._statusLabel.set_text(text);
    }
});

const TrashMenuButton = new Lang.Class({
    Name: 'PanelManager.TrashMenuButton',
    Extends: PanelMenu.Button,

    _init: function() {
        this.parent(0.0, _("Trash"), false);
        this.menu.actor.add_style_class_name('dschaerf-menu');
        this._hbox = new St.BoxLayout({ style_class: 'panel-status-menu-box' });
        this._icon = new St.Icon({ icon_name: 'user-trash-symbolic', style_class: 'system-status-icon' });
        this._hbox.add_child(this._icon);
        this._hbox.add_child(PopupMenu.arrowIcon(St.Side.BOTTOM));
        this.actor.add_child(this._hbox);
        this._itemTrashOpen = new TrashMenuItem(_("Open Trash"), 'user-trash-full');
        this._itemTrashOpen.connect('activate', Lang.bind(this, this._onOpenTrash));
        this.menu.addMenuItem(this._itemTrashOpen);
        this._itemTrashEmpty = new TrashMenuItem(_("Empty the Trash…"), 'user-trash');
        this._itemTrashEmpty.connect('activate', Lang.bind(this, this._onEmptyTrash));
        this.menu.addMenuItem(this._itemTrashEmpty);
        this.menu.addMenuItem(new PopupMenu.PopupSeparatorMenuItem());
        this._itemTrashContent = new TrashSubMenuMenuItem(_("Content of the Trash"), 'user-trash-full');
        this.menu.addMenuItem(this._itemTrashContent);
        this._trashMonitor = TRASH_URI.monitor_directory(0, null);
        this._trashMonitorChangedId = this._trashMonitor.connect('changed', Lang.bind(this, this._onTrashChanged));
        this._openStateChangedId = this.menu.connect('open-state-changed', Lang.bind(this, function(menu, isOpen) {
            if (isOpen) this._updateTrashContent();
        }));
        this._onTrashChanged();
        this.actor.connect('destroy', Lang.bind(this, function() {
            this.menu.disconnect(this._openStateChangedId);
            this._trashMonitor.disconnect(this._trashMonitorChangedId);
        }));
    },

    _onEmptyTrash: function() {
        let dialog = new UserDialog.ModalUserDialog('TrashEmpty');
        dialog.connect('trash-empty', Lang.bind(this, function() {
            let trashChild = null;
            let trashChildren = TRASH_URI.enumerate_children('*', 0, null);
            while ((trashChild = trashChildren.next_file(null)) !== null) {
                TRASH_URI.get_child(trashChild.get_name()).delete(null);
            }
            trashChildren.close(null);
            this.actor.hide();
        }));
        dialog.connect('trash-open', Lang.bind(this, this._onOpenTrash));
        dialog.open();
        Main.overview.hide();
    },

    _onOpenTrash: function() {
        Gio.AppInfo.launch_default_for_uri(TRASH_PATH, global.create_app_launch_context(0, -1));
        Main.overview.hide();
    },

    _onTrashChanged: function() {
        let info = TRASH_URI.query_info('standard::icon', 0, null);
        let iconName = info.get_icon().get_names()[0];
        this.actor.visible = iconName === 'user-trash-full' ? true : false;
    },

    _updateTrashContent: function() {
        this._itemTrashContent.menu.removeAll();
        this._itemTrashContent.setStatus('');
        let trashChild = null;
        let trashChildren = TRASH_URI.enumerate_children('*', 0, null);
        let trashEntrys = [];
        let count = 0;
        while ((trashChild = trashChildren.next_file(null)) !== null) {
            count++;
            if (count > TRASH_CONTENT_MAX) {
                this._itemTrashContent.setStatus(_("(only the first %s entries)").format(count - 1));
                break;
            }
            trashEntrys.push(trashChild);
        }
        if (trashEntrys.length > 0) trashEntrys.sort(function(a, b) {
            return GLib.utf8_collate(a.get_name(), b.get_name());
        });
        trashEntrys.forEach(function(trashEntry) {
            this._itemTrashContent.menu.addMenuItem(new TrashEntryMenuItem(trashEntry));
        }, this);
    }
});

function getExtension() {
    return Main.panel.addToStatusArea('dschaerfTrash', new TrashMenuButton(), 0, 'right');
}
