// -*- mode: js; js-indent-level: 4; indent-tabs-mode: nil -*-
/*
  Copyright (C) 2012-2019  Dieter Schärf <ds@dschaerf.de>

  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

const Gio = imports.gi.Gio;
const GLib = imports.gi.GLib;
const Lang = imports.lang;
const Mainloop = imports.mainloop;
const Meta = imports.gi.Meta;
const Shell = imports.gi.Shell;

const ExtensionUtils = imports.misc.extensionUtils;
const LoginManager = imports.misc.loginManager;
const Main = imports.ui.main;
const PopupMenu = imports.ui.popupMenu;
const StatusSystem = imports.ui.status.system;
const Util = imports.misc.util;

const Me = ExtensionUtils.getCurrentExtension();

const Convenience = Me.imports.convenience;
const UserDialog = Me.imports.userDialog;

const _ = Convenience.getTranslations();

const BASIC_SETTINGS = [
    'ca.desrt.dconf-editor.desktop',
    'gconf-editor.desktop',
    'gnome-control-center.desktop',
    'gnome-session-properties.desktop',
    'gnome-shell-extension-prefs.desktop',
    'gnome-tweak-tool.desktop',
    'org.gnome.tweaks.desktop',
    'system-config-printer.desktop',
    'synaptic.desktop',
    'users.desktop'
]

const DISABLE_RESTART_BUTTONS = 'disable-restart-buttons';

const GNOME_TERMINAL = [
    'gnome-terminal.desktop',
    'org.gnome.Terminal.desktop'
];

const SystemIndicator = new Lang.Class({
    Name: 'PanelManager.SystemMenu.SystemIndicator',

    _init: function() {
        this._appSystem = Shell.AppSystem.get_default();
        this._loginManager = LoginManager.getLoginManager();
        this._changedSystem(true);
        this._installedChangedId = this._appSystem.connect('installed-changed', Lang.bind(this, this._onInstalledChanged));
        this._openStateChangedId = Main.panel.statusArea.aggregateMenu._system.menu.connect('open-state-changed', Lang.bind(this, function(menu, isOpen) {
            if (!isOpen) return;
            Main.panel.statusArea.aggregateMenu._system._settingsAction.visible = !this._itemBasicSettings.actor.visible;
            this._updateHibernate();
            this._updateHybridSleep();
        }));
        this._onInstalledChanged();
    },

    _canHibernate: function(asyncCallback) {
        if (this._loginManager._proxy) {
            this._loginManager._proxy.call("CanHibernate", null, Gio.DBusCallFlags.NONE, -1, null, function(proxy, asyncResult) {
                try {
                    let result = proxy.call_finish(asyncResult).deep_unpack();
                    asyncCallback(result[0] !== 'no');
                } catch(e) {
                    asyncCallback(false);
                }
            });
        } else {
            Mainloop.idle_add(Lang.bind(this, function() {
                asyncCallback(false);
                return false;
            }));
        }
    },

    _canHybridSleep: function(asyncCallback) {
        if (this._loginManager._proxy) {
            this._loginManager._proxy.call("CanHybridSleep", null, Gio.DBusCallFlags.NONE, -1, null, function(proxy, asyncResult) {
                try {
                    let result = proxy.call_finish(asyncResult).deep_unpack();
                    asyncCallback(result[0] !== 'no');
                } catch(e) {
                    asyncCallback(false);
                }
            });
        } else {
            Mainloop.idle_add(Lang.bind(this, function() {
                asyncCallback(false);
                return false;
            }));
        }
    },

    _changedSystem: function(enabled) {
        if (enabled) {
            this._itemBasicSettings = new PopupMenu.PopupSubMenuMenuItem(_("Basic Settings"), true);
            this._itemBasicSettings.icon.icon_name = 'preferences-system-symbolic';
            Main.panel.statusArea.aggregateMenu._system.menu.addMenuItem(this._itemBasicSettings, 0);
            this._actionPanelManagerSettings = Main.panel.statusArea.aggregateMenu._system._createActionButton('preferences-other-symbolic', _("Settings for PanelManager"));
            this._actionPanelManagerSettings.connect('clicked', Lang.bind(this, this._onClickedPanelManagerSettings));
            this._actionGNOMEShellRestart = Main.panel.statusArea.aggregateMenu._system._createActionButton('view-refresh-symbolic', _("GNOME Shell restart"));
            this._actionGNOMEShellRestart.connect('clicked', Lang.bind(this, this._onClickedGNOMEShellRestart));
            this._actionGNOMETerminal = Main.panel.statusArea.aggregateMenu._system._createActionButton('utilities-terminal-symbolic', _("Terminal"));
            this._actionGNOMETerminal.connect('clicked', Lang.bind(this, this._onClickedTerminal));
            this._actionHibernate = Main.panel.statusArea.aggregateMenu._system._createActionButton('document-save-symbolic', _("Hibernate"));
            this._actionHibernate.connect('clicked', Lang.bind(this, this._onClickedHibernate));
            this._actionHybridSleep = Main.panel.statusArea.aggregateMenu._system._createActionButton('document-save-as-symbolic', _("Hybrid-Sleep"));
            this._actionHybridSleep.connect('clicked', Lang.bind(this, this._onClickedHybridSleep));
            this._altSwitcher = new StatusSystem.AltSwitcher(this._actionHibernate, this._actionHybridSleep);
            let actions = [];
            let children = Main.panel.statusArea.aggregateMenu._system._actionsItem.actor.get_children();
            for (let i = 1; i < children.length; i++) {
                Main.panel.statusArea.aggregateMenu._system._actionsItem.actor.remove_actor(children[i]);
                actions.push(children[i]);
            }
            for (let i = 0; i < actions.length; i++) {
                if (i === 0) Main.panel.statusArea.aggregateMenu._system._actionsItem.actor.add(this._actionPanelManagerSettings, { expand: true, x_fill: false });
                if (i === actions.length - 1) Main.panel.statusArea.aggregateMenu._system._actionsItem.actor.add(this._altSwitcher.actor, { expand: true, x_fill: false });
                Main.panel.statusArea.aggregateMenu._system._actionsItem.actor.add(actions[i], { expand: true, x_fill: false });
                if (i === 1) {
                    Main.panel.statusArea.aggregateMenu._system._actionsItem.actor.add(this._actionGNOMETerminal, { expand: true, x_fill: false });
                    Main.panel.statusArea.aggregateMenu._system._actionsItem.actor.add(this._actionGNOMEShellRestart, { expand: true, x_fill: false });
                }
            }
            if (Convenience.checkVersion([ '3.18', '3.20', '3.22', '3.24' ])) {
                Main.panel.statusArea.aggregateMenu._system._updateOrientationLock.call(Main.panel.statusArea.aggregateMenu._system);
                Main.panel.statusArea.aggregateMenu._system._updateLockScreen.call(Main.panel.statusArea.aggregateMenu._system);
                Main.panel.statusArea.aggregateMenu._system._updatePowerOff.call(Main.panel.statusArea.aggregateMenu._system);
                Main.panel.statusArea.aggregateMenu._system._updateSuspend.call(Main.panel.statusArea.aggregateMenu._system);
            } else {
                Main.panel.statusArea.aggregateMenu._system._systemActions._updateOrientationLock.call(Main.panel.statusArea.aggregateMenu._system._systemActions);
                Main.panel.statusArea.aggregateMenu._system._systemActions._updateLockScreen.call(Main.panel.statusArea.aggregateMenu._system._systemActions);
                Main.panel.statusArea.aggregateMenu._system._systemActions._updatePowerOff.call(Main.panel.statusArea.aggregateMenu._system._systemActions);
                Main.panel.statusArea.aggregateMenu._system._systemActions._updateSuspend.call(Main.panel.statusArea.aggregateMenu._system._systemActions);
            }
        } else {
            Main.panel.statusArea.aggregateMenu._system._actionsItem.actor.remove_actor(this._altSwitcher.actor);
            Main.panel.statusArea.aggregateMenu._system._actionsItem.actor.remove_actor(this._actionGNOMEShellRestart);
            Main.panel.statusArea.aggregateMenu._system._actionsItem.actor.remove_actor(this._actionGNOMETerminal);
            Main.panel.statusArea.aggregateMenu._system._actionsItem.actor.remove_actor(this._actionPanelManagerSettings);
            Main.panel.statusArea.aggregateMenu._system._settingsAction.show();
            this._actionHybridSleep.destroy();
            this._actionHibernate.destroy();
            this._altSwitcher.actor.destroy();
            this._actionGNOMEShellRestart.destroy();
            this._actionGNOMETerminal.destroy();
            this._actionPanelManagerSettings.destroy();
            this._itemBasicSettings.destroy();
        }
    },

    _onClickedGNOMEShellRestart: function() {
        Main.panel.statusArea.aggregateMenu._system.menu.itemActivated();
        Meta.restart(_("Restarting…"));
    },

    _onClickedHibernate: function() {
        Main.panel.statusArea.aggregateMenu._system.menu.itemActivated();
        this._dialog = new UserDialog.ModalUserDialog('Hibernate');
        this._dialog.connect('hybrid-sleep', Lang.bind(this, this._onClickedHybridSleep));
        this._dialog.connect('hibernate', Lang.bind(this, function() {
            if (this._loginManager._proxy) {
                this._loginManager._proxy.call("Hibernate", GLib.Variant.new('(b)', [true]), Gio.DBusCallFlags.NONE, -1, null, null);
            } else {
                this._loginManager.emit('prepare-for-sleep', true);
                this._loginManager.emit('prepare-for-sleep', false);
            }
        }));
        this._dialog.open();
        Main.overview.hide();
    },

    _onClickedHybridSleep: function() {
        Main.panel.statusArea.aggregateMenu._system.menu.itemActivated();
        if (this._loginManager._proxy) {
            this._loginManager._proxy.call("HybridSleep", GLib.Variant.new('(b)', [true]), Gio.DBusCallFlags.NONE, -1, null, null);
        } else {
            this._loginManager.emit('prepare-for-sleep', true);
            this._loginManager.emit('prepare-for-sleep', false);
        }
    },

    _onClickedPanelManagerSettings: function() {
        Main.panel.statusArea.aggregateMenu._system.menu.itemActivated();
        Util.spawnCommandLine('gnome-shell-extension-prefs '+Me.metadata['uuid']);
        Main.overview.hide();
    },

    _onClickedTerminal: function() {
        Main.panel.statusArea.aggregateMenu._system.menu.itemActivated();
        Util.spawnCommandLine('gnome-terminal');
        Main.overview.hide();
    },

    _onInstalledChanged: function() {
        this._itemBasicSettings.menu.removeAll();
        let appList = [];
        BASIC_SETTINGS.forEach(function(desktopFile) {
            let app = this._appSystem.lookup_app(desktopFile);
            if (app) appList.push(app);
        }, this);
        appList.sort(function(a, b) {
            return a.compare_by_name(b);
        });
        appList.forEach(function(app) {
            let item = new PopupMenu.PopupMenuItem(app.get_name());
            item.connect('activate', Lang.bind(this, function() {
                app.activate();
                Main.overview.hide();
            }));
            this._itemBasicSettings.menu.addMenuItem(item);
        }, this);
        this._itemBasicSettings.actor.visible = this._itemBasicSettings.menu.numMenuItems > 0;
        let gnomeTerminal = false;
        GNOME_TERMINAL.forEach(function(desktopFile) {
            let app = this._appSystem.lookup_app(desktopFile);
            if (app) gnomeTerminal = true;
        }, this);
        this._actionGNOMETerminal.visible = gnomeTerminal;
    },

    _updateHibernate: function() {
        this._canHibernate(Lang.bind(this, function(result) {
            let disableRestartButtons;
            let suspendNeedsAuth = false;
            if (Convenience.checkVersion([ '3.18', '3.20', '3.22', '3.24' ])) {
                disableRestartButtons = Main.panel.statusArea.aggregateMenu._system._loginScreenSettings.get_boolean(DISABLE_RESTART_BUTTONS);
                if (Convenience.checkVersion([ '3.22', '3.24' ])) suspendNeedsAuth = Main.panel.statusArea.aggregateMenu._system._suspendNeedsAuth;
            } else {
                disableRestartButtons = Main.panel.statusArea.aggregateMenu._system._systemActions._loginScreenSettings.get_boolean(DISABLE_RESTART_BUTTONS);
                suspendNeedsAuth = Main.panel.statusArea.aggregateMenu._system._systemActions._suspendNeedsAuth;
            }
            let disabled = (Main.sessionMode.isLocked && suspendNeedsAuth) || (Main.sessionMode.isGreeter && disableRestartButtons);
            this._actionHybridSleep.visible = result;
        }));
    },

    _updateHybridSleep: function() {
        this._canHybridSleep(Lang.bind(this, function(result) {
            let disableRestartButtons;
            let suspendNeedsAuth = false;
            if (Convenience.checkVersion([ '3.18', '3.20', '3.22', '3.24' ])) {
                disableRestartButtons = Main.panel.statusArea.aggregateMenu._system._loginScreenSettings.get_boolean(DISABLE_RESTART_BUTTONS);
                if (Convenience.checkVersion([ '3.22', '3.24' ])) suspendNeedsAuth = Main.panel.statusArea.aggregateMenu._system._suspendNeedsAuth;
            } else {
                disableRestartButtons = Main.panel.statusArea.aggregateMenu._system._systemActions._loginScreenSettings.get_boolean(DISABLE_RESTART_BUTTONS);
                suspendNeedsAuth = Main.panel.statusArea.aggregateMenu._system._systemActions._suspendNeedsAuth;
            }
            let disabled = (Main.sessionMode.isLocked && suspendNeedsAuth) || (Main.sessionMode.isGreeter && disableRestartButtons);
            this._actionHibernate.visible = result;
        }));
    },

    destroy: function() {
        Main.panel.statusArea.aggregateMenu._system.menu.disconnect(this._openStateChangedId);
        this._appSystem.disconnect(this._installedChangedId);
        this._changedSystem(false);
    }
});

function getExtension() {
    return new SystemIndicator();
}
