// -*- mode: js; js-indent-level: 4; indent-tabs-mode: nil -*-
/*
  Copyright (C) 2012-2019  Dieter Schärf <ds@dschaerf.de>

  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program. If not, see <http://www.gnu.org/licenses/>.
*/


const Gtk = imports.gi.Gtk;
const Gvc = imports.gi.Gvc;
const Lang = imports.lang;

const ExtensionUtils = imports.misc.extensionUtils;
const Main = imports.ui.main;
const PopupMenu = imports.ui.popupMenu;

const Me = ExtensionUtils.getCurrentExtension();

const Convenience = Me.imports.convenience;

const VOLUME_ICON_HIGHT = {
    _output: 'audio-volume-high-symbolic',
    _input:  'microphone-sensitivity-high-symbolic'
};
const VOLUME_ICON_LOW = {
    _output: 'audio-volume-low-symbolic',
    _input:  'microphone-sensitivity-low-symbolic'
};
const VOLUME_ICON_MEDIUM = {
    _output: 'audio-volume-medium-symbolic',
    _input:  'microphone-sensitivity-medium-symbolic'
};
const VOLUME_ICON_MUTED = {
    _output: 'audio-volume-muted-symbolic',
    _input:  'microphone-sensitivity-muted-symbolic'
};

const VolumeMenuItem = new Lang.Class({
    Name: 'PanelManger.VolumeMenuItem',
    Extends: PopupMenu.PopupSwitchMenuItem,

    _init: function(control, id) {
        this._control = control;
        this._stream = this._control.lookup_stream_id(id);
        this.parent(this._stream.description, !this._stream.is_muted);
        this._streamChangedId = this._stream.connect('notify::is-muted', Lang.bind(this, function() {
            this.setToggleState(!this._stream.is_muted);
        }));
        this.connect('toggled', Lang.bind(this, function() {
            this._stream.change_is_muted(!this._stream.is_muted);
            if (this._switch.state) {
                this._stream.volume = this._control.get_vol_max_norm();
                this._stream.push_volume();
            }
        }));
        this.actor.connect('destroy', Lang.bind(this, function() {
            this._stream.disconnect(this._streamChangedId);
        }));
    },

    _findHDMI: function(stream, iconNames) {
        let hdmi = false
        if (stream.get_ports && stream.get_ports().length > 0) hdmi = stream.get_port().port.indexOf('hdmi') >= 0;
        if (hdmi) iconNames.unshift('video-display', 'display');
    },

    _findHeadphone: function(stream, iconNames) {
        let headphone = false;
        if (stream.get_form_factor && (stream.get_form_factor() === 'headphone' || stream.get_form_factor() === 'headset')) headphone = true;
        if (stream.get_ports && stream.get_ports().length > 0) headphone = stream.get_port().port.indexOf('headphone') >= 0;
        if (headphone) iconNames.unshift('audio-headphones', 'headphones');
    },

    _findMicrophone: function(stream, iconNames) {
        let microphone = false;
        if (stream.get_ports && stream.get_ports().length > 0) microphone = stream.get_port().port.indexOf('mic') >= 0;
        if (microphone) iconNames.unshift('audio-input-microphone');
    },

    _findWebcam: function(stream, iconNames) {
        let webcam = false;
        if (stream.get_form_factor && stream.get_form_factor() === 'webcam') webcam = true;
        if (webcam) iconNames.unshift('camera-web-usb', 'camera-web', 'camera');
    },

    getIcon: function(stream) {
        let iconNames = [];
        if (stream.get_gicon && stream.get_gicon().get_names) iconNames = stream.get_gicon().get_names();
        this._findHDMI(stream, iconNames);
        this._findHeadphone(stream, iconNames);
        this._findMicrophone(stream, iconNames);
        this._findWebcam(stream, iconNames);
        if (iconNames.length < 1) iconNames = [ 'audio-card-pci', 'audio-card', 'audio' ];
        for (let i = 0; i < iconNames.length; i++) {
            let iconName = iconNames[i]+'-symbolic';
            if (Gtk.IconTheme.get_default().has_icon(iconName)) return iconName;
        }
        return '';
    },

    setSelected: function(value) {
        this.setOrnament(value ? PopupMenu.Ornament.CHECK : PopupMenu.Ornament.NONE);
    }
});

const VolumeSubMenuMenuItem = new Lang.Class({
    Name: 'PanelManger.VolumeSubMenuMenuItem',
    Extends: PopupMenu.PopupSubMenuMenuItem,

    _init: function(control) {
        this.parent('...', true);
        this._control = control;
    },

    setSelectedItem: function(item) {
        this.label.set_text(item.label.get_text());
        this.icon.icon_name = item.getIcon(item._stream);
        if (this['selectedItem']) this.selectedItem.setSelected(false);
        this.selectedItem = item;
        this.selectedItem.setSelected(true);
    },

    addVolumeMenuItem: function(stream, getDefaultStream, setDefaultStream) {
        let item = new VolumeMenuItem(this._control, stream.id);
        this.menu.addMenuItem(item);
        if (stream.id === item._control[getDefaultStream]().id) this.setSelectedItem(item);
        item.connect('activate', Lang.bind(this, function() {
            if (item._stream.id !== item._control[getDefaultStream]().id) {
                item._control[setDefaultStream](item._stream);
                item._stream.change_is_muted(false);
                item.setToggleState(true);
                this.setSelectedItem(item);
            }
            if (this.menu.isOpen) this.menu.close(0);
            let topMenu = item._getTopMenu();
            if (!topMenu.isOpen) topMenu.open(-1);
        }));
    }
});

const VolumeIndicator = new Lang.Class({
    Name: 'PanelManager.SystemMenu.VolumeIndicator',

    _init: function() {
        this._inputIndicator = Main.panel.statusArea.aggregateMenu._volume._addIndicator();
        this._inputDevices = new VolumeSubMenuMenuItem(Main.panel.statusArea.aggregateMenu._volume._control);
        Main.panel.statusArea.aggregateMenu._volume._volumeMenu.addMenuItem(this._inputDevices, 2);
        this._outputDevices = new VolumeSubMenuMenuItem(Main.panel.statusArea.aggregateMenu._volume._control);
        Main.panel.statusArea.aggregateMenu._volume._volumeMenu.addMenuItem(this._outputDevices, 1);
        this._controlChangedId = [
            Main.panel.statusArea.aggregateMenu._volume._control.connect('state-changed', Lang.bind(this, this._onControlStateChanged)),
            Main.panel.statusArea.aggregateMenu._volume._control.connect('default-sink-changed', Lang.bind(this, this._readOutput)),
            Main.panel.statusArea.aggregateMenu._volume._control.connect('default-source-changed', Lang.bind(this, this._readInput)),
            Main.panel.statusArea.aggregateMenu._volume._control.connect('stream-added', Lang.bind(this, this._updateVisibility)),
            Main.panel.statusArea.aggregateMenu._volume._control.connect('stream-removed', Lang.bind(this, this._updateVisibility))
        ];
        this._iconChangedId = Main.panel.statusArea.aggregateMenu._volume._volumeMenu.connect('icon-changed', Lang.bind(this, this._updateOutputIndicator));
        this._onControlStateChanged();
    },

    _getIcon: function(property) {
        if (!this[property] || this[property].is_muted || this[property].volume <= 0) {
            return VOLUME_ICON_MUTED[property];
        } else {
            let n = Math.floor(3 * this[property].volume / Main.panel.statusArea.aggregateMenu._volume._control.get_vol_max_norm()) + 1;
            if (n < 2) return VOLUME_ICON_LOW[property];
            if (n >= 3) return VOLUME_ICON_HIGHT[property];
            return VOLUME_ICON_MEDIUM[property];
        }
    },

    _onControlStateChanged: function() {
        this._outputDevices.menu.removeAll();
        this._inputDevices.menu.removeAll();
        if (Main.panel.statusArea.aggregateMenu._volume._control.get_state() === Gvc.MixerControlState.READY) {
            let sinks = Main.panel.statusArea.aggregateMenu._volume._control.get_sinks();
            if (sinks.length > 0) {
                sinks.sort(function(a, b) {
                    return a.description.toLowerCase() > b.description.toLowerCase();
                });
                sinks.forEach(function(sink) {
                    this._outputDevices.addVolumeMenuItem(sink, 'get_default_sink', 'set_default_sink');
                }, this);
                this._readOutput();
            }
            let sources = Main.panel.statusArea.aggregateMenu._volume._control.get_sources();
            if (sources.length > 0) {
                sources.sort(function(a, b) {
                    return a.description.toLowerCase() > b.description.toLowerCase();
                });
                sources.forEach(function(source) {
                    this._inputDevices.addVolumeMenuItem(source, 'get_default_source', 'set_default_source');
                }, this);
                this._readInput();
            }
        } else {
            this._showInput = false;
            this._showOutput = false;
            this._updateVisibility();
        }
    },

    _readInput: function() {
        this._showInput = false;
        if (this._inputChangedId) this._inputChangedId.forEach(function(id) {
            this._input.disconnect(id);
        }, this);
        this._input = Main.panel.statusArea.aggregateMenu._volume._control.get_default_source();
        if (this._input) {
            this._inputChangedId = [
                this._input.connect('notify::port', Lang.bind(this, this._updateInputSliderIcon)),
                this._input.connect('notify::is-muted', Lang.bind(this, this._updateInputIndicator)),
                this._input.connect('notify::volume', Lang.bind(this, this._updateInputIndicator))
            ];
            let menuItems = this._inputDevices.menu._getMenuItems();
            menuItems.forEach(function(item) {
                if (item._stream.id === this._input.id) this._inputDevices.setSelectedItem(item);
            }, this);
            this._showInput = true;
            this._updateInputIndicator();
            this._updateInputSliderIcon();
        }
        this._updateVisibility();
    },

    _readOutput: function() {
        this._showOutput = false;
        if (this._outputChangedId) this._output.disconnect(this._outputChangedId);
        this._output = Main.panel.statusArea.aggregateMenu._volume._control.get_default_sink();
        if (this._output) {
            this._outputChangedId = this._output.connect('notify::port', Lang.bind(this, this._updateOutputSliderIcon));
            let menuItems = this._outputDevices.menu._getMenuItems();
            menuItems.forEach(function(item) {
                if (item._stream.id === this._output.id) this._outputDevices.setSelectedItem(item);
            }, this);
            this._showOutput = true;
            this._updateOutputIndicator();
            this._updateOutputSliderIcon();
        }
        this._updateVisibility();
    },

    _updateInputIndicator: function() {
        this._inputIndicator.icon_name = this._getIcon('_input');
    },

    _updateInputSliderIcon: function() {
        Main.panel.statusArea.aggregateMenu._volume._volumeMenu._input._icon.icon_name = 'audio-input-microphone-symbolic';
    },

    _updateOutputIndicator: function() {
        Main.panel.statusArea.aggregateMenu._volume._primaryIndicator.icon_name = this._getIcon('_output');
    },

    _updateOutputSliderIcon: function() {
        Main.panel.statusArea.aggregateMenu._volume._volumeMenu._output._icon.icon_name = 'audio-speakers-symbolic';
    },

    _updateVisibility: function() {
        Main.panel.statusArea.aggregateMenu._volume.indicators.visible = this._showOutput || this._showInput;
        Main.panel.statusArea.aggregateMenu._volume._primaryIndicator.visible = this._showOutput;
        this._inputIndicator.visible = this._showInput;
        Main.panel.statusArea.aggregateMenu._volume.menu.actor.visible = this._showOutput || this._showInput;
        Main.panel.statusArea.aggregateMenu._volume._volumeMenu._output.item.actor.visible = this._showOutput;
        this._outputDevices.actor.visible = this._showOutput && this._outputDevices.menu.numMenuItems > 0;
        Main.panel.statusArea.aggregateMenu._volume._volumeMenu._input.item.actor.visible = this._showInput;
        this._inputDevices.actor.visible = this._showInput && this._inputDevices.menu.numMenuItems > 0;
    },

    destroy: function() {
        if (this._inputChangedId) this._inputChangedId.forEach(function(id) {
            this._input.disconnect(id);
        }, this);
        if (this._outputChangedId) this._output.disconnect(this._outputChangedId);
        Main.panel.statusArea.aggregateMenu._volume._volumeMenu.disconnect(this._iconChangedId);
        this._controlChangedId.forEach(function(id) {
            Main.panel.statusArea.aggregateMenu._volume._control.disconnect(id);
        }, this);
        this._updateVisibility();
        this._inputDevices.destroy();
        this._outputDevices.destroy();
        this._inputIndicator.destroy();
        Main.panel.statusArea.aggregateMenu._volume._volumeMenu._output._updateSliderIcon();
    }
});

function getExtension() {
    return new VolumeIndicator();
}
